# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import sys

import tobiko
from tobiko.common.managers import loader
from tobiko.tests.unit import TobikoUnitTest


SOME_NONE = None


class SomeClass(object):

    def some_method(self):
        pass


class TestLoader(TobikoUnitTest):

    def setUp(self):
        super(TestLoader, self).setUp()
        self.manager = loader.LoaderManager()
        self.patch(loader, 'LOADERS', self.manager)

    def test_load_object_with_none(self):
        object_id = '.'.join([__name__, 'SOME_NONE'])
        obj = tobiko.load_object(object_id)
        self.assertIsNone(obj)

        _loader = self.manager.get_loader(object_id)
        self.assertEqual(_loader.id, object_id)
        self.assertFalse(_loader.is_module)
        self.assertIs(_loader.get(), obj)
        self.assertIs(_loader, self.manager.get_loader(object_id))

    def test_load_module_with_none(self):
        object_id = '.'.join([__name__, 'SOME_NONE'])
        module = tobiko.load_module(object_id)
        self.assertEqual(__name__, module.__name__)

    def test_load_object_with_module(self):
        object_id = __name__
        obj = tobiko.load_object(object_id)
        self.assertIs(sys.modules[object_id], obj)

        _loader = self.manager.get_loader(object_id)
        self.assertEqual(_loader.id, object_id)
        self.assertTrue(_loader.is_module)
        self.assertIs(_loader.get(), obj)
        self.assertIs(_loader, self.manager.get_loader(object_id))

    def test_load_module_with_module(self):
        object_id = __name__
        module = tobiko.load_module(object_id)
        self.assertEqual(__name__, module.__name__)

    def test_load_object_with_class(self):
        object_id = '.'.join([SomeClass.__module__,
                              SomeClass.__name__])
        obj = tobiko.load_object(object_id)
        self.assertIs(SomeClass, obj)

    def test_load_module_with_class(self):
        object_id = '.'.join([SomeClass.__module__,
                              SomeClass.__name__])
        module = tobiko.load_module(object_id)
        self.assertEqual(__name__, module.__name__)

    def test_load_object_with_class_method(self):
        object_id = '.'.join([SomeClass.__module__,
                              SomeClass.__name__,
                              SomeClass.some_method.__name__])
        obj = tobiko.load_object(object_id)
        self.assertEqual(SomeClass.some_method, obj)

        _loader = self.manager.get_loader(object_id)
        self.assertEqual(_loader.id, object_id)
        self.assertFalse(_loader.is_module)
        self.assertIs(_loader.get(), obj)
        self.assertIs(_loader, self.manager.get_loader(object_id))

    def test_load_module_with_class_method(self):
        object_id = '.'.join([SomeClass.__module__,
                              SomeClass.__name__,
                              SomeClass.some_method.__name__])
        module = tobiko.load_module(object_id)
        self.assertEqual(__name__, module.__name__)

    def test_load_object_with_non_existing(self):
        object_id = '.'.join([SomeClass.__module__, '<non-existing>'])
        self.assertRaises(ImportError, tobiko.load_object, object_id)

    def test_load_module_with_non_existing(self):
        object_id = '.'.join([SomeClass.__module__, '<non-existing>'])
        self.assertRaises(ImportError, tobiko.load_module, object_id)

    def test_load_object_with_non_existing_member(self):
        object_id = '.'.join([SomeClass.__module__,
                              SomeClass.__name__,
                              '<non-existing>'])
        self.assertRaises(AttributeError, tobiko.load_object, object_id)

    def test_load_module_with_non_existing_member(self):
        object_id = '.'.join([SomeClass.__module__,
                              SomeClass.__name__,
                              '<non-existing>'])
        self.assertRaises(AttributeError, tobiko.load_module, object_id)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr

import tobiko
from tobiko.openstack.neutron import _client
from tobiko.shell import ssh
from tobiko.shell import ping


NeutronPortType = typing.Dict[str, typing.Any]


def list_port_ip_addresses(port: NeutronPortType,
                           subnet_id: typing.Optional[str] = None,
                           ip_version: typing.Optional[int] = None,
                           check_connectivity: bool = False,
                           ssh_client: ssh.SSHClientFixture = None) -> \
        tobiko.Selection[netaddr.IPAddress]:
    addresses = tobiko.Selection[netaddr.IPAddress](
        netaddr.IPAddress(fixed_ip['ip_address'])
        for fixed_ip in port['fixed_ips']
        if subnet_id is None or subnet_id == fixed_ip['subnet_id'])
    if ip_version:
        addresses = addresses.with_attributes(version=ip_version)
    if addresses and check_connectivity:
        hosts = ping.list_reachable_hosts(addresses, ssh_client=ssh_client)
        addresses = tobiko.Selection(netaddr.IPAddress(host) for host in hosts)
    return addresses


def find_port_ip_address(port: NeutronPortType, unique: bool = False,
                         **kwargs) -> netaddr.IPAddress:
    addresses = list_port_ip_addresses(port=port, **kwargs)
    if unique:
        return addresses.unique
    else:
        return addresses.first


def list_device_ip_addresses(device_id: str,
                             network_id: typing.Optional[str] = None,
                             ip_version: typing.Optional[int] = None,
                             check_connectivity: bool = False,
                             ssh_client: ssh.SSHClientFixture = None,
                             need_dhcp: typing.Optional[bool] = None,
                             **subnet_params) -> \
        tobiko.Selection[netaddr.IPAddress]:
    ports = _client.list_ports(device_id=device_id,
                               network_id=network_id)
    if need_dhcp is not None:
        subnet_params['enable_dhcp'] = bool(need_dhcp)
    subnets = _client.list_subnets(network_id=network_id,
                                   ip_version=ip_version,
                                   **subnet_params)
    addresses = tobiko.Selection[netaddr.IPAddress](
        port_ip
        for subnet in subnets
        for port in ports
        for port_ip in list_port_ip_addresses(port=port,
                                              subnet_id=subnet['id'],
                                              ip_version=ip_version))
    if addresses and check_connectivity:
        hosts = ping.list_reachable_hosts(addresses, ssh_client=ssh_client)
        addresses = tobiko.Selection(netaddr.IPAddress(host) for host in hosts)
    return addresses


def find_device_ip_address(device_id: str, unique: bool = False, **kwargs):
    addresses = list_device_ip_addresses(device_id=device_id,  **kwargs)
    if unique:
        return addresses.unique
    else:
        return addresses.first



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing
import json

from oslo_log import log

import tobiko
from tobiko.openstack.neutron import _agent


LOG = log.getLogger(__name__)


def wait_for_master_and_backup_agents(
        router_id: str,
        unique_master: bool = True,
        timeout: tobiko.Seconds = None,
        interval: tobiko.Seconds = None) -> \
        typing.Tuple[typing.Dict, typing.List[typing.Dict]]:
    for attempt in tobiko.retry(timeout=timeout,
                                interval=interval,
                                default_timeout=300.,
                                default_interval=5.):
        router_agents = _agent.list_l3_agent_hosting_routers(router_id)
        master_agents = router_agents.with_items(ha_state='active')
        if master_agents:
            LOG.debug(
                f"Router '{router_id}' has {len(master_agents)} master "
                "agent(s):\n"
                f"{json.dumps(master_agents, indent=4, sort_keys=True)}")
        backup_agents = router_agents.with_items(ha_state='standby')
        if backup_agents:
            LOG.debug(
                f"Router '{router_id}' has {len(backup_agents)} backup "
                "agent(s)):\n"
                f"{json.dumps(backup_agents, indent=4, sort_keys=True)}")
        other_agents = [agent
                        for agent in router_agents
                        if (agent not in master_agents + backup_agents)]
        if other_agents:
            LOG.debug(
                f"Router '{router_id}' has {len(other_agents)} other "
                "agent(s):\n"
                f"{json.dumps(master_agents, indent=4, sort_keys=True)}")
        try:
            if unique_master:
                master_agent = master_agents.unique
            else:
                master_agent = master_agents.first
        except (tobiko.MultipleObjectsFound, tobiko.ObjectNotFound):
            attempt.check_limits()
        else:
            break
    else:
        raise RuntimeError("tobiko retry loop ended before break?")

    return master_agent, backup_agents



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import inspect
import typing
from unittest import mock

import tobiko
from tobiko.tests import unit


@tobiko.protocol
class MyProto:
    # pylint: disable=unused-argument

    def call_one(self, arg='a') -> int:
        return 42

    def call_two(self, *args) -> int:
        return 42

    def call_three(self, **kwargs) -> int:
        return 42


class MyProtoHandler(tobiko.CallHandler):

    protocol_class = MyProto


class ProxyTest(unit.TobikoUnitTest):

    def handle_call(self, method: typing.Callable, *_args, **_kwargs):
        self.assertTrue(inspect.isfunction(method))

    def mock_handler(self) -> typing.Callable:
        return typing.cast(typing.Callable,
                           mock.MagicMock(side_effect=self.handle_call))

    def test_call_handler(self):
        # pylint: disable=no-member
        handler = self.mock_handler()
        proxy: MyProto = MyProtoHandler(handler).use_as(MyProto)
        self.assertIsInstance(proxy, MyProto)
        self.assertTrue(callable(proxy.call_one))
        self.assertEqual(inspect.signature(MyProto.call_one),
                         inspect.signature(type(proxy).call_one))
        self.assertIsNot(MyProto.call_one,
                         proxy.call_one)
        proxy.call_one()
        handler.assert_called_with(MyProto.call_one, 'a')

    def test_call_proxy(self):
        handler = self.mock_handler()
        proxy = tobiko.call_proxy(MyProto, handler).use_as(MyProto)
        self.assertIsInstance(proxy, MyProto)
        self.assertTrue(callable(proxy.call_one))
        self.assertEqual(inspect.signature(MyProto.call_one),
                         inspect.signature(type(proxy).call_one))
        self.assertIsNot(MyProto.call_one,
                         proxy.call_one)
        proxy.call_one()
        handler.assert_called_with(MyProto.call_one, 'a')
        proxy.call_one('b')
        handler.assert_called_with(MyProto.call_one, 'b')
        proxy.call_one(arg='c')
        handler.assert_called_with(MyProto.call_one, 'c')

        proxy.call_two(1, 2, 3)
        handler.assert_called_with(MyProto.call_two, 1, 2, 3)

        proxy.call_three(a='a', b='b')
        handler.assert_called_with(MyProto.call_three, a='a', b='b')



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import socket
import typing

import docker
from docker.transport import unixconn
from docker import constants

import tobiko
from tobiko.docker import _exception
from tobiko.docker import _shell
from tobiko.shell import ssh


def get_docker_client(base_urls: typing.List[str] = None,
                      ssh_client: ssh.SSHClientType = None,
                      sudo=False):
    return DockerClientFixture(base_urls=base_urls,
                               ssh_client=ssh_client,
                               sudo=sudo)


def list_docker_containers(client=None, **kwargs):
    try:
        containers = docker_client(client).containers.list(all=True,
                                                           sparse=True,
                                                           **kwargs)
    except _exception.DockerUrlNotFoundError:
        return tobiko.Selection()
    else:
        return tobiko.select(containers)


def docker_client(obj=None):
    if obj is None:
        obj = get_docker_client()
    if tobiko.is_fixture(obj):
        obj = tobiko.setup_fixture(obj).client
    if isinstance(obj, docker.DockerClient):
        return obj
    raise TypeError('Cannot obtain a DockerClient from {!r}'.format(obj))


class DockerClientFixture(tobiko.SharedFixture):

    base_urls: typing.Optional[typing.List[str]] = None
    ssh_client = ssh.SSHClientType = None
    sudo = False
    client: typing.Optional['DockerClient'] = None

    def __init__(self,
                 base_urls: typing.Iterable[str] = None,
                 ssh_client: ssh.SSHClientType = None,
                 sudo: bool = None):
        super(DockerClientFixture, self).__init__()
        if base_urls:
            self.base_urls = list(base_urls)
        if ssh_client is not None:
            self.ssh_client = ssh_client
        if sudo is not None:
            self.sudo = sudo

    def setup_fixture(self):
        self.setup_ssh_client()
        self.setup_base_urls()
        self.setup_client()

    def setup_ssh_client(self):
        ssh_client = self.ssh_client
        if ssh_client is None:
            self.ssh_client = ssh_client = ssh.ssh_proxy_client() or False
            if ssh_client:
                tobiko.setup_fixture(ssh_client)
        return ssh_client

    def setup_base_urls(self):
        base_urls = self.base_urls
        if base_urls is None:
            self.base_urls = base_urls = self.discover_docker_urls()
        return base_urls

    def setup_client(self):
        client = self.client
        if client is None:
            self.client = client = self.create_client()
        return client

    def create_client(self):
        error: typing.Optional[Exception] = None
        for base_url in self.base_urls:
            client = DockerClient(base_url=base_url,
                                  ssh_client=self.ssh_client,
                                  sudo=self.sudo)
            try:
                client.ping()
            except Exception as ex:
                ex.__cause__ = error
                error = ex
            else:
                return client

        if isinstance(error, Exception):
            raise error
        else:
            raise _exception.DockerError('Unable to create docker client')

    def connect(self):
        return tobiko.setup_fixture(self).client

    def discover_docker_urls(self):
        return _shell.discover_docker_urls(ssh_client=self.ssh_client,
                                           sudo=self.sudo)


class DockerClient(docker.DockerClient):

    def __init__(self,
                 base_url: str,
                 ssh_client: ssh.SSHClientType = None,
                 sudo=False):
        # pylint: disable=super-init-not-called
        self.api = APIClient(base_url=base_url,
                             ssh_client=ssh_client,
                             sudo=sudo)


class APIClient(docker.APIClient):

    def __init__(self,
                 base_url: str,
                 ssh_client: ssh.SSHClientType = None,
                 sudo=False):
        self.ssh_client = ssh_client
        self.sudo = sudo
        super(APIClient, self).__init__(base_url=base_url)

    def get_adapter(self, url):
        adapter = super(APIClient, self).get_adapter(url)
        if isinstance(adapter, unixconn.UnixHTTPAdapter):
            new_adapter = UnixHTTPAdapter(
                socket_url=f"http+unix://{adapter.socket_path}",
                timeout=adapter.timeout,
                max_pool_size=adapter.max_pool_size,
                ssh_client=self.ssh_client,
                sudo=self.sudo)
            self._custom_adapter = new_adapter
            for prefix, other_adapter in self.adapters.items():
                if adapter is other_adapter:
                    self.adapters[prefix] = new_adapter
            return new_adapter
        else:
            return adapter


class UnixHTTPAdapter(unixconn.UnixHTTPAdapter):

    def __init__(self,
                 socket_url: str,
                 timeout=60,
                 pool_connections=constants.DEFAULT_NUM_POOLS,
                 max_pool_size=constants.DEFAULT_MAX_POOL_SIZE,
                 ssh_client: ssh.SSHClientType = None,
                 sudo=False):
        self.ssh_client = ssh_client
        self.sudo = sudo
        super().__init__(socket_url=socket_url,
                         timeout=timeout,
                         pool_connections=pool_connections,
                         max_pool_size=max_pool_size)

    def get_connection(self, url, proxies=None):
        with self.pools.lock:
            pool = self.pools.get(url)
            if pool:
                return pool

            pool = UnixHTTPConnectionPool(base_url=url,
                                          socket_path=self.socket_path,
                                          timeout=self.timeout,
                                          maxsize=self.max_pool_size,
                                          ssh_client=self.ssh_client,
                                          sudo=self.sudo)
            self.pools[url] = pool

        return pool


class UnixHTTPConnectionPool(unixconn.UnixHTTPConnectionPool):

    def __init__(self,
                 base_url: str,
                 socket_path: str,
                 timeout=60,
                 maxsize=10,
                 ssh_client: ssh.SSHClientType = None,
                 sudo=False):
        self.ssh_client = ssh_client
        self.sudo = sudo
        super().__init__(base_url=base_url, socket_path=socket_path,
                         timeout=timeout, maxsize=maxsize)

    def _new_conn(self):
        return UnixHTTPConnection(base_url=self.base_url,
                                  unix_socket=self.socket_path,
                                  timeout=self.timeout,
                                  ssh_client=self.ssh_client,
                                  sudo=self.sudo)


class UnixHTTPConnection(unixconn.UnixHTTPConnection):

    def __init__(self,
                 base_url: str,
                 unix_socket: str,
                 timeout=60,
                 ssh_client: ssh.SSHClientType = None,
                 sudo=False):
        self.ssh_client = ssh_client
        self.sudo = sudo
        super(UnixHTTPConnection, self).__init__(base_url=base_url,
                                                 unix_socket=unix_socket,
                                                 timeout=timeout)

    def connect(self):
        ssh_client = ssh.ssh_client_fixture(self.ssh_client)
        if ssh_client is None:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.settimeout(self.timeout)
            sock.connect(self.unix_socket)
        elif self.sudo:
            client = self.ssh_client.connect()
            sock = client.get_transport().open_session(timeout=self.timeout)
            command = f"sudo nc -U '{self.unix_socket}'"
            sock.exec_command(command)
        else:
            sock = self.ssh_client.open_unix_socket(
                socket_path=self.unix_socket)
        self.sock = sock



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import time

from oslo_log import log
import netaddr

import tobiko
from tobiko.shell import sh


LOG = log.getLogger(__name__)
CURL_OPTIONS = "-f --connect-timeout 2 -g"


def request(client_stack, ip_address, protocol, port, ssh_client=None):
    ssh_client = ssh_client or client_stack.ssh_client

    if netaddr.IPAddress(ip_address) == 6:
        ip_address = f"[{ip_address}]"

    cmd = f"curl {CURL_OPTIONS} {protocol.lower()}://{ip_address}:{port}/id"

    return sh.ssh_execute(ssh_client, cmd).stdout


def check_members_balanced(pool_stack, client_stack,
                           members_count,
                           loadbalancer_vip, loadbalancer_protocol,
                           loadbalancer_port, ssh_client=None):

    """Check if traffic is properly balanced between members."""

    test_case = tobiko.get_test_case()

    replies = {}

    for _ in range(members_count * 10):
        content = request(
            client_stack, loadbalancer_vip,
            loadbalancer_protocol, loadbalancer_port, ssh_client)

        if content not in replies:
            replies[content] = 0
        replies[content] += 1

        # wait one second (required when using cirros' nc fake webserver)
        time.sleep(1)

    LOG.debug("Replies from load balancer: {}".format(replies))

    # assert that 'members_count' servers replied
    test_case.assertEqual(members_count, len(replies),
                          'The number of detected active members:{} is not '
                          'as expected:{}'.format(len(replies), members_count))

    if pool_stack.lb_algorithm == 'ROUND_ROBIN':
        # assert that requests have been fairly dispatched (each server
        # received the same number of requests)
        test_case.assertEqual(1, len(set(replies.values())),
                              'The number of requests served by each member is'
                              ' different and not as expected by used '
                              'ROUND_ROBIN algorithm.')



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import octavia
from tobiko.openstack import stacks


@keystone.skip_if_missing_service(name='octavia')
class OctaviaBasicTrafficScenarioTest(testtools.TestCase):
    """Octavia traffic scenario test.

    Create a load balancer with 2 members that run a server application,
    Create a client that is connected to the load balancer VIP port,
    Generate network traffic from the client to the load balanacer.
    """
    loadbalancer_stack = tobiko.required_setup_fixture(
        stacks.OctaviaLoadbalancerStackFixture)

    listener_stack = tobiko.required_setup_fixture(
        stacks.OctaviaListenerStackFixture)

    pool_stack = tobiko.required_setup_fixture(
        stacks.OctaviaPoolStackFixture)

    member1_stack = tobiko.required_setup_fixture(
        stacks.OctaviaMemberServerStackFixture)

    member2_stack = tobiko.required_setup_fixture(
        stacks.OctaviaOtherMemberServerStackFixture)

    client_stack = tobiko.required_setup_fixture(
        stacks.OctaviaClientServerStackFixture)

    members_count = 2

    def setUp(self):
        # pylint: disable=no-member
        super(OctaviaBasicTrafficScenarioTest, self).setUp()

        self.loadbalancer_vip = self.loadbalancer_stack.loadbalancer_vip
        self.loadbalancer_port = self.listener_stack.lb_port
        self.loadbalancer_protocol = self.listener_stack.lb_protocol

        octavia.wait_for_status(status_key=octavia.PROVISIONING_STATUS,
                                status=octavia.ACTIVE,
                                get_client=octavia.get_member,
                                object_id=self.pool_stack.pool_id,
                                member_id=self.member1_stack.member_id)

        octavia.wait_for_status(status_key=octavia.PROVISIONING_STATUS,
                                status=octavia.ACTIVE,
                                get_client=octavia.get_member,
                                object_id=self.pool_stack.pool_id,
                                member_id=self.member2_stack.member_id)

        # Wait for LB is provisioned and ACTIVE
        octavia.wait_for_status(status_key=octavia.PROVISIONING_STATUS,
                                status=octavia.ACTIVE,
                                get_client=octavia.get_loadbalancer,
                                object_id=(
                                    self.loadbalancer_stack.loadbalancer_id))

    @property
    def loadbalancer(self):
        return self.loadbalancer_stack

    def test_traffic(self):
        octavia.check_members_balanced(self.pool_stack,
                                       self.client_stack,
                                       self.members_count,
                                       self.loadbalancer_vip,
                                       self.loadbalancer_protocol,
                                       self.loadbalancer_port)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.openstack import tests
from tobiko.shell import ping


class ServerStackResourcesTest(testtools.TestCase):

    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def test_server(self):
        "Test actual server details"
        server = self.stack.resources.server
        self.assertEqual('OS::Nova::Server', server.resource_type)
        # Verify actual server status (is alive, etc.)
        nova_client = nova.get_nova_client()
        server_data = nova_client.servers.get(server.physical_resource_id)
        self.assertEqual(self.stack.server_name, server_data.name)

    def test_port(self):
        "Test actual server port details"
        port = self.stack.resources.port
        self.assertEqual('OS::Neutron::Port', port.resource_type)
        # Verify actual port status (is alive, etc.)
        port_data = neutron.get_port(port.physical_resource_id)
        self.assertEqual(port.physical_resource_id, port_data['id'])


class ServerCreationTest(testtools.TestCase):

    def setUp(self):
        testtools.TestCase.setUp(self)
        if config.get_bool_env('TOBIKO_PREVENT_CREATE'):
            self.skip('TOBIKO_PREVENT_CREATE is true')

    def test_server_creation(self):
        tests.test_server_creation()

    def test_evacuable_server_creation(self):
        tests.test_evacuable_server_creation()

    def test_server_creation_and_shutoff(self):
        tests.test_server_creation_and_shutoff()

    def test_servers_creation(self):
        tests.test_servers_creation()


class MigrateServerStack(stacks.CirrosServerStackFixture):
    pass


@keystone.skip_unless_has_keystone_credentials()
@nova.skip_if_missing_hypervisors(count=2)
class MigrateServerTest(testtools.TestCase):

    stack = tobiko.required_setup_fixture(MigrateServerStack)
    peer_stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def test_migrate_server(self, live=False):
        """Tests cold migration actually changes hypervisor
        """
        server = self.setup_server()
        initial_hypervisor = nova.get_server_hypervisor(server)

        server = self.migrate_server(server, live=live)

        final_hypervisor = nova.get_server_hypervisor(server)
        self.assertNotEqual(initial_hypervisor, final_hypervisor)

    def test_migrate_server_with_host(self, live=False):
        """Tests cold migration actually ends on target hypervisor
        """
        server = self.setup_server()
        initial_hypervisor = nova.get_server_hypervisor(server)
        for hypervisor in nova.list_hypervisors(status='enabled', state='up'):
            if initial_hypervisor != hypervisor.hypervisor_hostname:
                target_hypervisor = hypervisor.hypervisor_hostname
                break
        else:
            target_hypervisor = None
            self.skipTest("Cannot find a valid hypervisor host to migrate "
                          "server to")

        server = self.migrate_server(server=server, host=target_hypervisor,
                                     live=live)

        final_hypervisor = nova.get_server_hypervisor(server)
        self.assertEqual(target_hypervisor, final_hypervisor)

    @tobiko.skip("Expected to create problems on compute nodes")
    def test_live_migrate_server(self):
        self.test_migrate_server(live=True)

    @tobiko.skip("Expected to create problems on compute nodes")
    def test_live_migrate_server_with_host(self):
        self.test_migrate_server_with_host(live=True)

    def setup_server(self):
        server = self.stack.ensure_server_status('ACTIVE')
        self.assertEqual('ACTIVE', server.status)
        return server

    def migrate_server(self, server, live=False, **params):
        self.assertEqual('ACTIVE', server.status)
        # wait until all VM fixed IPs are reachable
        ping.assert_reachable_hosts(self.stack.list_fixed_ips(),
                                    timeout=900.,
                                    ssh_client=self.peer_stack.ssh_client)

        if live:
            nova.live_migrate_server(server, **params)
            server = nova.wait_for_server_status(
                server, 'ACTIVE', transient_status=['MIGRATING'])
        else:
            nova.migrate_server(server, **params)

            server = nova.wait_for_server_status(server, 'VERIFY_RESIZE')
            self.assertEqual('VERIFY_RESIZE', server.status)
            nova.confirm_resize(server)

            server = nova.wait_for_server_status(
                server, 'ACTIVE', transient_status=['VERIFY_RESIZE'])
        self.assertEqual('ACTIVE', server.status)

        # wait until all VM fixed IPs are reachable
        ping.assert_reachable_hosts(self.stack.list_fixed_ips(),
                                    timeout=900.,
                                    ssh_client=self.peer_stack.ssh_client)
        # check floating IP is reachable
        ping.assert_reachable_hosts([self.stack.floating_ip_address])
        return server



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import time

from heatclient.v1 import client as heatclient
from heatclient import exc
import mock
import yaml

import tobiko
from tobiko.openstack import heat
from tobiko.openstack.heat import _stack
from tobiko.openstack import keystone
from tobiko.tests.unit import openstack


class MyStack(heat.HeatStackFixture):
    template = heat.heat_template({'template': 'from-class'})


class MyStackWithStackName(MyStack):
    stack_name = 'stack.name.from.class'


class MyStackWithParameters(MyStack):
    parameters = heat.heat_stack_parameters({'param': 'from-class'})


class MyStackWithWaitInterval(MyStack):
    wait_interval = 10


class MyTemplateFixture(heat.HeatTemplateFixture):
    template = {'template': 'from-class'}


class MockClient(mock.NonCallableMagicMock):
    pass


class HeatStackFixtureTest(openstack.OpenstackTest):

    def setUp(self):
        super(HeatStackFixtureTest, self).setUp()
        self.patch(heatclient, 'Client', MockClient)

    def test_init(self, fixture_class=MyStack, stack_name=None,
                  template=None, parameters=None, wait_interval=None,
                  client=None):
        stack = fixture_class(stack_name=stack_name, template=template,
                              parameters=parameters,
                              wait_interval=wait_interval, client=client)

        if stack_name:
            self.assertEqual(stack_name, stack.stack_name)
        elif isinstance(fixture_class.stack_name, property):
            self.assertEqual(tobiko.get_fixture_name(stack), stack.stack_name)
        else:
            self.assertEqual(fixture_class.stack_name, stack.stack_name)

        self.check_stack_template(stack=stack, template=template)

        self.assertIsInstance(stack.parameters,
                              _stack.HeatStackParametersFixture)
        self.assertEqual(parameters or getattr(fixture_class.parameters,
                                               'parameters', {}),
                         stack.parameters.parameters)
        self.assertEqual(wait_interval or fixture_class.wait_interval,
                         stack.wait_interval)
        self.assertIs(client or fixture_class.client, stack.client)

    def test_init_with_stack_name(self):
        self.test_init(stack_name='my-stack-name')

    def test_init_with_stack_name_from_class(self):
        self.test_init(fixture_class=MyStackWithStackName)

    def test_init_with_template(self):
        self.test_init(template={'some': 'template'})

    def test_init_with_template_fixture(self):
        self.test_init(template=MyTemplateFixture())

    def test_init_with_parameters(self):
        self.test_init(parameters={'my': 'value'})

    def test_init_with_parameters_from_class(self):
        self.test_init(fixture_class=MyStackWithParameters)

    def test_init_with_wait_interval(self):
        self.test_init(wait_interval=20)

    def test_init_with_wait_interval_from_class(self):
        self.test_init(fixture_class=MyStackWithWaitInterval)

    def test_init_with_client(self):
        session = keystone.get_keystone_session()
        self.test_init(client=heatclient.Client(session=session))

    def test_init_with_client_fixture(self):
        self.test_init(client=heat.HeatClientFixture())

    def test_setup(self, fixture_class=MyStack, template=None,
                   stack_name=None, parameters=None, wait_interval=None,
                   stacks=None, create_conflict=False, call_create=True,
                   call_delete=False, call_sleep=False):
        client = MockClient()
        stacks = stacks or [
            exc.HTTPNotFound,
            mock_stack('CREATE_IN_PROGRESS')]
        client.stacks.get.side_effect = stacks

        if create_conflict:
            client.stacks.create.side_effect = exc.HTTPConflict
        else:
            client.stacks.create.return_value = {
                'stack': {'id': '<stack-id>'}}

        sleep = self.patch(time, 'sleep')
        stack = fixture_class(stack_name=stack_name, parameters=parameters,
                              template=template, wait_interval=wait_interval,
                              client=client)

        stack.setUp()

        self.assertIs(client, stack.client)
        self.assertEqual(wait_interval or fixture_class.wait_interval,
                         stack.wait_interval)
        client.stacks.get.assert_has_calls([mock.call(stack.stack_name,
                                                      resolve_outputs=False)])

        if call_delete:
            client.stacks.delete.assert_called_once_with(stack.stack_id)
        else:
            client.stacks.delete.assert_not_called()

        parameters = (parameters or
                      (fixture_class.parameters and
                       fixture_class.parameters.values) or
                      {})
        self.assertEqual(parameters, stack.parameters.values)
        if call_create:
            client.stacks.create.assert_called_once_with(
                parameters=parameters, stack_name=stack.stack_name,
                template=yaml.safe_dump(stack.template.template))
        else:
            client.stacks.create.assert_not_called()

        if call_sleep:
            sleep.assert_called()

    def test_setup_with_stack_name(self):
        self.test_setup(stack_name='my-stack-name')

    def test_setup_with_stack_name_from_class(self):
        self.test_setup(fixture_class=MyStackWithStackName)

    def test_setup_with_template(self):
        self.test_setup(template={'other': 'template'})

    def test_setup_with_template_fixture(self):
        self.test_setup(template=heat.heat_template({'template':
                                                     'from-fixture'}))

    def test_setup_with_template_fixture_type(self):
        self.test_setup(template=MyTemplateFixture)

    def test_setup_with_parameters(self):
        self.test_setup(parameters={'from_init': True})

    def test_setup_with_parameters_from_class(self):
        self.test_setup(fixture_class=MyStackWithParameters)

    def test_setup_with_wait_interval(self):
        self.test_setup(wait_interval=10.)

    def test_setup_with_wait_interval_from_class(self):
        self.test_setup(fixture_class=MyStackWithWaitInterval)

    def test_setup_when_none(self):
        self.test_setup(stacks=[None,
                                mock_stack('CREATE_IN_PROGRESS')])

    def test_setup_when_delete_completed(self):
        self.test_setup(stacks=[mock_stack('DELETE_COMPLETE'),
                                None,
                                mock_stack('CREATE_IN_PROGRESS')])

    def test_setup_when_delete_failed(self):
        self.test_setup(stacks=[mock_stack('DELETE_FAILED'),
                                mock_stack('DELETE_IN_PROGRESS'),
                                mock_stack('DELETE_COMPLETE'),
                                None,
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_delete=True, call_sleep=True)

    def test_setup_when_delete_failed_fast_delete(self):
        self.test_setup(stacks=[mock_stack('DELETE_FAILED'),
                                None,
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_delete=True)

    def test_setup_when_create_complete(self):
        self.test_setup(stacks=[mock_stack('CREATE_COMPLETE')],
                        call_create=False)

    def test_setup_when_create_failed(self):
        self.test_setup(stacks=[mock_stack('CREATE_FAILED'),
                                mock_stack('DELETE_IN_PROGRESS'),
                                mock_stack('DELETE_COMPLETE'),
                                None,
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_delete=True, call_sleep=True)

    def test_setup_when_create_failed_fast_delete(self):
        self.test_setup(stacks=[mock_stack('CREATE_FAILED'),
                                None,
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_delete=True)

    def test_setup_when_create_in_progress(self):
        self.test_setup(stacks=[mock_stack('CREATE_IN_PROGRESS')],
                        call_create=False)

    def test_setup_when_delete_in_progress_then_complete(self):
        self.test_setup(stacks=[mock_stack('DELETE_IN_PROGRESS'),
                                mock_stack('DELETE_COMPLETE'),
                                None,
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_sleep=True)

    def test_setup_when_delete_in_progress_then_failed(self):
        self.test_setup(stacks=[mock_stack('DELETE_IN_PROGRESS'),
                                mock_stack('DELETE_FAILED'),
                                mock_stack('DELETE_COMPLETE'),
                                None,
                                mock_stack('CREATE_IN_PROGRESS')],
                        call_sleep=True, call_delete=True)

    def test_setup_when_create_conflict(self):
        self.test_setup(create_conflict=True)

    def test_setup_when_prevent_to_create_set(self):
        with mock.patch.dict('os.environ', {'TOBIKO_PREVENT_CREATE': 'True'}):
            self.test_setup(stacks=[mock_stack('CREATE_COMPLETE')],
                            call_create=False)

    def test_cleanup(self):
        client = MockClient()
        client.stacks.get.return_value = None
        stack = MyStack(client=client)
        stack.cleanUp()
        client.stacks.delete.assert_called_once_with(stack.stack_name)

    def test_outputs(self):
        stack = mock_stack(status='CREATE_COMPLETE',
                           outputs=[{'output_key': 'key1',
                                     'output_value': 'value1'},
                                    {'output_key': 'key2',
                                     'output_value': 'value2'}])
        client = MockClient()
        client.stacks.get.return_value = stack
        stack_fixture = MyStack(
            template={'outputs': {'key1': {}, 'key2': {}}},
            client=client)

        outputs = stack_fixture.outputs

        self.assertEqual('value1', outputs.key1)
        self.assertEqual('value2', outputs.key2)

    def test_parameters(self):
        stack_fixture = MyStack(
            template={'parameters': {'key1': {}, 'key2': {}}},
            parameters={'key1': 'value1',
                        'key2': 'value2'})

        parameters = stack_fixture.parameters

        self.assertEqual('value1', parameters.key1)
        self.assertEqual('value2', parameters.key2)

    def check_stack_template(self, stack, template):
        expected_template = template or type(stack).template
        if tobiko.is_fixture(expected_template):
            self.assertIs(expected_template, stack.template)
        elif isinstance(expected_template, collections.Mapping):
            self.assertEqual(expected_template, stack.template.template)
        else:
            message = "Unsupported template type: {!r}".format(
                expected_template)
            self.fail(message)


def mock_stack(status, stack_id='<stack-id>', outputs=None):
    return mock.MagicMock(stack_status=status,
                          id=stack_id,
                          outputs=outputs or [])



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import typing

from oslo_log import log

from tobiko.openstack.topology import _topology
from tobiko.shell import ping


LOG = log.getLogger(__name__)


def assert_reachable_nodes(
        nodes: typing.Iterable[_topology.OpenStackTopologyNode],
        **ping_params):
    node_ips = {node.name: str(node.public_ip) for node in nodes}
    LOG.debug(f"Test nodes are reachable: "
              f"{json.dumps(node_ips, sort_keys=True, indent=4)}")
    ping.assert_reachable_hosts(node_ips.values(), **ping_params)


def assert_unreachable_nodes(
        nodes: typing.Iterable[_topology.OpenStackTopologyNode],
        **ping_params):
    node_ips = {node.name: str(node.public_ip) for node in nodes}
    LOG.debug(f"Test nodes are unreachable: "
              f"{json.dumps(node_ips, sort_keys=True, indent=4)}")
    ping.assert_unreachable_hosts(node_ips.values(), **ping_params)



#!/usr/bin/env python3
# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import site
import sys
import typing


TOP_DIR = os.path.realpath(os.path.dirname(os.path.dirname(__file__)))

if TOP_DIR not in sys.path:
    sys.path.insert(0, TOP_DIR)

from tools import common  # noqa
from tools import get_version  # noqa

LOG = common.get_logger(__name__)

TOX_VERSION = os.environ.get('TOX_VERSION') or '>=3.8.0'

TOX_CONSTRAINTS_FILE = (
    os.environ.get('TOX_CONSTRAINTS_FILE') or
    'https://opendev.org/openstack/requirements/raw/branch/master/upper-constraints.txt')

TOX_CONSTRAINTS = (
    os.environ.get('TOX_CONSTRAINTS') or f"'-c{TOX_CONSTRAINTS_FILE}'")


def main():
    common.setup_logging()
    install_tox()
    install_bindeps()
    install_podman1()
    install_tobiko()


def install_tox(version=TOX_VERSION):
    LOG.info(f"Installing Tox... (version: {version})")
    pip_install(f"'tox{version}'")


def install_bindeps():
    LOG.info(f"Installing Tobiko binary dependencies...")
    common.execute(os.path.join(TOP_DIR, 'tools', 'install-bindeps.sh'),
                   capture_stdout=False)


def install_tobiko():
    version = get_version.get_version()
    LOG.info(f"Installing Tobiko version {version}...")
    pip_install(f"-e '{TOP_DIR}'")


def install_podman1(version='===1.6.0'):
    pip_unisntall('podman')

    LOG.info(f"Installing Podman... (version: {version})")

    site_dirs = {os.path.dirname(os.path.realpath(site_dir))
                 for site_dir in site.getsitepackages()
                 if os.path.isdir(site_dir)}
    more_site_dirs = {os.path.join(site_dir, 'site-packages')
                      for site_dir in site_dirs
                      if os.path.isdir(os.path.join(site_dir, 'site-packages'))}
    site_dirs.update(more_site_dirs)
    LOG.debug(f"Site packages dirs: {site_dirs}")

    # Must ensure pre-existing podman directories are restored
    # after installation
    podman_dirs = [os.path.join(site_dir, 'podman')
                   for site_dir in sorted(site_dirs)]
    LOG.debug(f"Possible podman directories: {podman_dirs}")
    with common.stash_dir(*podman_dirs):
        for podman_dir in podman_dirs:
            assert not os.path.exists(podman_dir)
        pip_install(f"'podman{version}'")
        for podman_dir in podman_dirs:
            if os.path.isdir(podman_dir):
                # Rename podman directory to podman1
                os.rename(podman_dir, podman_dir + '1')
                break
        else:
            raise RuntimeError("Podman directory not found!")
        for podman_dir in podman_dirs:
            assert not os.path.exists(podman_dir)


def pip_install(args):
    LOG.debug(f"Installing packages: {args}...")
    common.execute_python(f"-m pip install {TOX_CONSTRAINTS} {args}",
                          capture_stdout=False)


def pip_unisntall(args):
    LOG.debug(f"Uninstalling packages: {args}...")
    common.execute_python(f"-m pip uninstall -y {args}",
                          capture_stdout=False)


if __name__ == '__main__':
    main()



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr
import testtools

import tobiko
from tobiko.shell import curl
from tobiko.shell import ssh
from tobiko.openstack import keystone
from tobiko.openstack import stacks


@keystone.skip_unless_has_keystone_credentials()
class TestCurl(testtools.TestCase):

    stack = tobiko.required_setup_fixture(stacks.UbuntuServerStackFixture)

    def test_execute_curl(
            self,
            ip_address: typing.Optional[netaddr.IPAddress] = None,
            ssh_client: typing.Optional[ssh.SSHClientFixture] = None):
        if ip_address is None:
            # Use the floating IP
            ip_address = self.stack.ip_address
        http_port = self.stack.http_port
        result = curl.execute_curl(scheme='http',
                                   hostname=ip_address,
                                   port=http_port,
                                   path='/id',
                                   ssh_client=ssh_client,
                                   connect_timeout=10.,
                                   retry_count=30,
                                   retry_timeout=300.,
                                   retry_interval=10.).strip()
        self.assertEqual(self.stack.server_name, result)

    def test_execute_curl_ipv4(self):
        self.test_execute_curl(ip_address=self.get_fixed_ip(ip_version=4),
                               ssh_client=self.stack.ssh_client)

    def test_execute_curl_ipv6(self):
        self.test_execute_curl(ip_address=self.get_fixed_ip(ip_version=6),
                               ssh_client=self.stack.ssh_client)

    def get_fixed_ip(self, ip_version):
        for fixed_ip in self.stack.fixed_ips:
            ip_address = netaddr.IPAddress(fixed_ip['ip_address'])
            if ip_version == ip_address.version:
                return ip_address
        self.skipTest(f"Server {self.stack.server_id} has any "
                      f"IPv{ip_version} address.")



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import ast
import re
import typing

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack.topology import _topology
from tobiko.shell import files


class NeutronNovaResponse(typing.NamedTuple):
    hostname: str
    name: str
    server_uuid: str
    status: str
    code: int
    tag: str
    line: str


class NeutronNovaResponseReader(tobiko.SharedFixture):
    log_digger: files.MultihostLogFileDigger
    groups = ['controller']
    message_pattern = r'Nova event response: '
    service_name = neutron.SERVER
    responses: tobiko.Selection[NeutronNovaResponse]

    def setup_fixture(self):
        self.log_digger = self.useFixture(
            _topology.get_log_file_digger(
                service_name=self.service_name,
                groups=self.groups,
                pattern=self.message_pattern))
        self.read_responses()

    def read_responses(self) \
            -> tobiko.Selection[NeutronNovaResponse]:
        responses = tobiko.Selection[NeutronNovaResponse]()
        message_pattern = re.compile(self.message_pattern)
        for hostname, line in self.log_digger.find_lines(
                new_lines=hasattr(self, 'responses')):
            found = message_pattern.search(line)
            assert found is not None
            response_text = line[found.end():].strip()
            response_data = ast.literal_eval(response_text)
            assert isinstance(response_data, dict)
            response = NeutronNovaResponse(
                hostname=hostname,
                line=line,
                **response_data)
            responses.append(response)
        if hasattr(self, 'responses'):
            self.responses.extend(responses)
        else:
            self.responses = responses
        return responses


def read_neutron_nova_responses(
        reader: NeutronNovaResponseReader = None,
        new_lines=True,
        **attributes) \
        -> tobiko.Selection[NeutronNovaResponse]:
    if reader is None:
        reader = tobiko.setup_fixture(NeutronNovaResponseReader)
    responses = reader.read_responses()
    if not new_lines:
        responses = reader.responses
    if attributes and responses:
        responses = responses.with_attributes(**attributes)
    return responses



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from keystoneclient import discover
from keystoneclient.v2_0 import client as client_v2
from keystoneclient.v3 import client as client_v3
from oslo_log import log

from tobiko.openstack import keystone
from tobiko.tests.unit import openstack
from tobiko.tests.unit.openstack import test_client

LOG = log.getLogger(__name__)
KEYSTONE_CLIENTS = client_v2.Client, client_v3.Client


class DiscoverMock(object):

    def __init__(self, session, **kwargs):
        self.session = session
        self.kwargs = kwargs

    def create_client(self, version, unstable):
        LOG.debug("Create a mock keystone client for version %r "
                  "(unestable=%r)", version, unstable)
        return client_v3.Client(session=self.session)


class KeystoneClientFixtureTest(test_client.OpenstackClientFixtureTest):

    def setUp(self):
        super(KeystoneClientFixtureTest, self).setUp()
        self.patch(discover, 'Discover', DiscoverMock)

    def create_client(self, session=None):
        return keystone.KeystoneClientFixture(session=session)


class GetKeystoneClientTest(openstack.OpenstackTest):

    def setUp(self):
        super(GetKeystoneClientTest, self).setUp()
        self.patch(discover, 'Discover', DiscoverMock)

    def test_get_keystone_client(self, session=None, shared=True):
        client1 = keystone.get_keystone_client(session=session, shared=shared)
        client2 = keystone.get_keystone_client(session=session, shared=shared)
        if shared:
            self.assertIs(client1, client2)
        else:
            self.assertIsNot(client1, client2)
        self.assertIsInstance(client1, KEYSTONE_CLIENTS)
        self.assertIsInstance(client2, KEYSTONE_CLIENTS)

    def test_get_keystone_client_with_not_shared(self):
        self.test_get_keystone_client(shared=False)

    def test_get_keystone_client_with_session(self):
        session = keystone.get_keystone_session()
        self.test_get_keystone_client(session=session)


class KeystoneClientTest(openstack.OpenstackTest):

    def setUp(self):
        super(KeystoneClientTest, self).setUp()
        self.patch(discover, 'Discover', DiscoverMock)

    def test_keystone_client_with_none(self):
        default_client = keystone.get_keystone_client()
        client = keystone.keystone_client(None)
        self.assertIsInstance(client, KEYSTONE_CLIENTS)
        self.assertIs(default_client, client)

    def test_keystone_client_with_client(self):
        default_client = keystone.get_keystone_client()
        client = keystone.keystone_client(default_client)
        self.assertIsInstance(client, KEYSTONE_CLIENTS)
        self.assertIs(default_client, client)

    def test_keystone_client_with_fixture(self):
        fixture = keystone.KeystoneClientFixture()
        client = keystone.keystone_client(fixture)
        self.assertIsInstance(client, KEYSTONE_CLIENTS)
        self.assertIs(client, fixture.client)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import os
import sys
import typing  # noqa

from oslo_log import log
import testtools
import yaml

import tobiko


LOG = log.getLogger(__name__)


class KeystoneCredentials(collections.namedtuple(
        'KeystoneCredentials', ['api_version',
                                'auth_url',
                                'username',
                                'password',
                                'project_name',
                                'domain_name',
                                'user_domain_name',
                                'project_domain_name',
                                'project_domain_id',
                                'cacert',
                                'trust_id'])):

    def to_dict(self):
        return {k: v
                for k, v in self._asdict().items()
                if v is not None}

    def __repr__(self):
        params = self.to_dict()
        if 'password' in params:
            params['password'] = '***'
        return 'keystone_credentials({!s})'.format(
            ", ".join("{!s}={!r}".format(k, v)
                      for k, v in sorted(params.items())))

    required_params = ('auth_url', 'username', 'password', 'project_name')

    def validate(self, required_params=None):
        required_params = required_params or self.required_params
        missing_params = [p
                          for p in required_params
                          if not getattr(self, p)]
        if missing_params:
            reason = "undefined parameters: {!s}".format(
                ', '.join(missing_params))
            raise InvalidKeystoneCredentials(credentials=self, reason=reason)


class NoSuchCredentialsError(tobiko.TobikoException):
    message = "No such credentials from any of: {fixtures}"


class KeystoneCredentialsFixture(tobiko.SharedFixture):

    credentials: typing.Optional[KeystoneCredentials] = None

    def __init__(self,
                 credentials: typing.Optional[KeystoneCredentials] = None):
        super(KeystoneCredentialsFixture, self).__init__()
        if credentials is not None:
            self.credentials = credentials

    def setup_fixture(self):
        self.setup_credentials()

    def setup_credentials(self):
        credentials = self.credentials
        if credentials is None:
            credentials = self.get_credentials()
            if credentials is not None:
                try:
                    credentials.validate()
                except InvalidKeystoneCredentials as ex:
                    LOG.info("No such valid credentials from %r (%r)",
                             self, ex)
                else:
                    self.addCleanup(self.cleanup_credentials)
                    self.credentials = credentials

    def cleanup_credentials(self):
        del self.credentials

    def get_credentials(self) -> typing.Optional[KeystoneCredentials]:
        return self.credentials


KeystoneCredentialsType = typing.Union[None,
                                       KeystoneCredentials,
                                       KeystoneCredentialsFixture,
                                       str,
                                       typing.Type]


def get_keystone_credentials(obj: KeystoneCredentialsType = None) -> \
        typing.Optional[KeystoneCredentials]:
    if obj is None:
        return default_keystone_credentials()
    if isinstance(obj, KeystoneCredentials):
        return obj
    if tobiko.is_fixture(obj):
        obj = tobiko.get_fixture(obj)
        if isinstance(obj, KeystoneCredentialsFixture):
            obj = tobiko.setup_fixture(obj).credentials
            return get_keystone_credentials(obj)
    raise TypeError(f"Can't get {KeystoneCredentials} object from {obj}")


def default_keystone_credentials() -> typing.Optional[KeystoneCredentials]:
    credentials = tobiko.setup_fixture(
        DefaultKeystoneCredentialsFixture).credentials
    if credentials is not None:
        tobiko.check_valid_type(credentials, KeystoneCredentials)
    return credentials


def keystone_credentials(api_version=None,
                         auth_url=None,
                         username=None,
                         password=None,
                         project_name=None,
                         domain_name=None,
                         user_domain_name=None,
                         project_domain_name=None,
                         project_domain_id=None,
                         cacert=None,
                         trust_id=None,
                         cls=KeystoneCredentials) -> KeystoneCredentials:
    return cls(api_version=api_version,
               auth_url=auth_url,
               username=username,
               password=password,
               project_name=project_name,
               domain_name=domain_name,
               user_domain_name=user_domain_name,
               project_domain_name=project_domain_name,
               project_domain_id=project_domain_id,
               cacert=cacert,
               trust_id=trust_id)


class InvalidKeystoneCredentials(tobiko.TobikoException):
    message = "invalid Keystone credentials; {reason!s}; {credentials!r}"


class EnvironKeystoneCredentialsFixture(KeystoneCredentialsFixture):

    environ: typing.Optional[typing.Dict[str, str]] = None

    def __init__(self,
                 credentials: typing.Optional[KeystoneCredentials] = None,
                 environ: typing.Optional[typing.Dict[str, str]] = None):
        super(EnvironKeystoneCredentialsFixture, self).__init__(
            credentials=credentials)
        if environ is not None:
            self.environ = environ

    def setup_fixture(self):
        if self.environ is None:
            self.environ = self.get_environ()
        super(EnvironKeystoneCredentialsFixture, self).setup_fixture()

    def get_environ(self) -> typing.Optional[typing.Dict[str, str]]:
        return dict(os.environ)

    def get_credentials(self) -> typing.Optional[KeystoneCredentials]:
        auth_url = self.get_env('OS_AUTH_URL')
        if not auth_url:
            LOG.debug("OS_AUTH_URL environment variable not defined")
            return None

        api_version = (
            self.get_int_env('OS_IDENTITY_API_VERSION') or
            api_version_from_url(auth_url))
        username = (
            self.get_env('OS_USERNAME') or
            self.get_env('OS_USER_ID'))
        password = self.get_env('OS_PASSWORD')
        project_name = (
            self.get_env('OS_PROJECT_NAME') or
            self.get_env('OS_TENANT_NAME') or
            self.get_env('OS_PROJECT_ID') or
            self.get_env('OS_TENANT_ID'))
        if api_version == 2:
            return keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=username,
                password=password,
                project_name=project_name)
        else:
            domain_name = (
                self.get_env('OS_DOMAIN_NAME') or
                self.get_env('OS_DOMAIN_ID'))
            user_domain_name = (
                self.get_env('OS_USER_DOMAIN_NAME') or
                self.get_env('OS_USER_DOMAIN_ID'))
            project_domain_name = (
                self.get_env('OS_PROJECT_DOMAIN_NAME'))
            project_domain_id = (
                self.get_env('OS_PROJECT_DOMAIN_ID'))
            trust_id = self.get_env('OS_TRUST_ID')
            return keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=username,
                password=password,
                project_name=project_name,
                domain_name=domain_name,
                user_domain_name=user_domain_name,
                project_domain_name=project_domain_name,
                project_domain_id=project_domain_id,
                trust_id=trust_id)

    def get_env(self, name) -> typing.Optional[str]:
        environ = self.environ
        if environ is None:
            return None
        else:
            return environ.get(name)

    def get_int_env(self, name) -> typing.Optional[int]:
        value = self.get_env(name=name)
        if value is None:
            return None
        else:
            return int(value)


def has_keystone_credentials(obj=None) -> bool:
    try:
        credentials = get_keystone_credentials(obj)
    except NoSuchCredentialsError:
        return False
    else:
        return credentials is not None


def skip_unless_has_keystone_credentials(*args, **kwargs):
    return tobiko.skip_unless('Missing Keystone credentials',
                              has_keystone_credentials, *args, **kwargs)


class ConfigKeystoneCredentialsFixture(KeystoneCredentialsFixture):

    def get_credentials(self) -> typing.Optional[KeystoneCredentials]:
        from tobiko import config
        conf = config.CONF.tobiko.keystone
        auth_url = conf.auth_url
        if not auth_url:
            LOG.debug("auth_url option not defined in 'keystone' section of "
                      "tobiko.conf")
            return None

        api_version = (conf.api_version or
                       api_version_from_url(auth_url))
        if api_version == 2:
            return keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=conf.username,
                password=conf.password,
                project_name=conf.project_name)
        else:
            return keystone_credentials(
                api_version=api_version,
                auth_url=auth_url,
                username=conf.username,
                password=conf.password,
                project_name=conf.project_name,
                domain_name=conf.domain_name,
                user_domain_name=conf.user_domain_name,
                project_domain_name=conf.project_domain_name,
                project_domain_id=conf.project_domain_id,
                trust_id=conf.trust_id)


DEFAULT_KEYSTONE_CREDENTIALS_FIXTURES = [
    EnvironKeystoneCredentialsFixture,
    ConfigKeystoneCredentialsFixture]


class DefaultKeystoneCredentialsFixture(KeystoneCredentialsFixture):

    fixtures = DEFAULT_KEYSTONE_CREDENTIALS_FIXTURES

    def get_credentials(self) -> typing.Optional[KeystoneCredentials]:
        errors = []
        for fixture in self.fixtures:
            try:
                credentials = tobiko.setup_fixture(fixture).credentials
            except Exception as ex:
                LOG.debug("Error getting credentials from %r: %s",
                          tobiko.get_fixture_name(fixture), ex)
                errors.append(tobiko.exc_info())
                continue

            if credentials:
                LOG.info("Got default credentials from fixture %r: %r",
                         fixture, credentials)
                return credentials
            else:
                LOG.debug('Got no credentials from %r', fixture)

        if len(errors) == 1:
            errors[0].reraise()
        elif errors:
            raise testtools.MultipleExceptions(errors)

        raise NoSuchCredentialsError(fixtures=self.fixtures)


def api_version_from_url(auth_url) -> typing.Optional[int]:
    if auth_url.endswith('/v2.0'):
        LOG.debug('Got Keystone API version 2 from auth_url: %r', auth_url)
        return 2
    elif auth_url.endswith('/v3'):
        LOG.debug('Got Keystone API version 3 from auth_url: %r', auth_url)
        return 3
    else:
        LOG.warning('Unable to get Keystone API version from auth_url:  %r',
                    auth_url)
        return None


def print_credentials():
    credentials = default_keystone_credentials()
    yaml.dump(dict(credentials.to_dict()),
              sys.stdout,
              indent=4,
              sort_keys=True)
